/* Copyright (c) 2006..2017 Dirk Jagdmann <doj@cubic.org>

   This software is provided 'as-is', without any express or implied
   warranty. In no event will the authors be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you
   must not claim that you wrote the original software. If you use
   this software in a product, an acknowledgment in the product
   documentation would be appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and
   must not be misrepresented as being the original software.

   3. This notice may not be removed or altered from any source
   distribution. */

/** @file
 * convert C++ integer literals to other bases.
 * See http://en.cppreference.com/w/cpp/language/integer_literal for details.
 */

#define _DEFAULT_SOURCE 1
#define _BSD_SOURCE 1
#include <endian.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string>

/**
 * convert a string with a binary literal to an integer.
 * @param s string with a binary literal.
 * @return integer value of s.
 * This program will exit() if the string s contains a character that is not a binary literal.
 */
intmax_t binary(std::string s)
{
    intmax_t x=0;
    if (s.find("0b") == 0) {
	s.erase(0,2);
    }
    for(auto c : s)
    {
	x*=2;
	if(c == '0') /* do nothing */ ;
	else if(c == '1') ++x;
	else
	{
	    printf("illegal character '%c' for binary numbers!\n", c);
	    exit(1);
	}
    }
    return x;
}

/**
 * check if the character c is a valid integer literal character.
 * @return true if c is a valid integer literal character.
 * @return false if c is not a valid integer literal character.
 */
bool isvalid(const char c)
{
    const char valid[] = {
	'-',
	'0','1','2','3','4','5','6','7','8','9',
	'a','b','c','d','e','f','x',
	'A','B','C','D','E','F','X',
    };
    for(unsigned i=0; i<sizeof(valid); ++i)
	if(valid[i] == c)
	    return true;
    return false;
}

/**
 * remove optional characters from s to prepare a string for integer literal parsing.
 * white space and the single quote character are removed.
 * @returned pruned s.
 */
std::string prune(const std::string s)
{
    std::string o;
    o.reserve(s.size());
    for(auto c : s)
    {
	if (::isspace(c)) continue;
	if (c == '\'') continue;
	o += c;
    }
    return o;
}

int main(int argc, char **argv)
{
    for(int j = 1; j < argc; ++j) {
	intmax_t i = 0;
	uintmax_t u = 0;
	const std::string arg = prune(argv[j]);

	if(arg.empty()) { continue; }
#define X(s) else if (arg == #s) { i = s; u = s; }
	X(INT8_MIN)
	    X(INT16_MIN)
	    X(INT32_MIN)
	    X(INT64_MIN)
	    X(INT8_MAX)
	    X(INT16_MAX)
	    X(INT32_MAX)
	    X(INT64_MAX)
	    X(UINT8_MAX)
	    X(UINT16_MAX)
	    X(UINT32_MAX)
	    X(UINT64_MAX)
	    X(INT_LEAST8_MIN)
	    X(INT_LEAST16_MIN)
	    X(INT_LEAST32_MIN)
	    X(INT_LEAST64_MIN)
	    X(INT_LEAST8_MAX)
	    X(INT_LEAST16_MAX)
	    X(INT_LEAST32_MAX)
	    X(INT_LEAST64_MAX)
	    X(UINT_LEAST8_MAX)
	    X(UINT_LEAST16_MAX)
	    X(UINT_LEAST32_MAX)
	    X(UINT_LEAST64_MAX)
	    X(INT_FAST8_MIN)
	    X(INT_FAST16_MIN)
	    X(INT_FAST32_MIN)
	    X(INT_FAST64_MIN)
	    X(INT_FAST8_MAX)
	    X(INT_FAST16_MAX)
	    X(INT_FAST32_MAX)
	    X(INT_FAST64_MAX)
	    X(UINT_FAST8_MAX)
	    X(UINT_FAST16_MAX)
	    X(UINT_FAST32_MAX)
	    X(UINT_FAST64_MAX)
	    X(INTPTR_MIN)
	    X(INTPTR_MAX)
	    X(UINTPTR_MAX)
	    X(INTMAX_MIN)
	    X(INTMAX_MAX)
	    X(UINTMAX_MAX)
	    X(PTRDIFF_MIN)
	    X(PTRDIFF_MAX)
	    X(SIG_ATOMIC_MIN)
	    X(SIG_ATOMIC_MAX)
	    X(SIZE_MAX)
	    X(WCHAR_MIN)
	    X(WCHAR_MAX)
	    X(WINT_MIN)
	    X(WINT_MAX)
#undef X
	else {
	    // check for valid characters
	    for(auto c : arg) {
		if(!isvalid(c)) {
		    printf("illegal character '%c' in %s!\n", c, arg.c_str());
		    goto nextarg;
		}
	    }
	    // parse arg
	    if (arg.find("0b") == 0) {
		i = u = binary(arg);
	    } else {
		i = strtoll(arg.c_str(),0,0);
		u = strtoull(arg.c_str(),0,0);
	    }
	}
	{
	    // print value in different formats
	    printf("dec = %+ji %ju le16:%u be16:%u le32:%u be32:%u le64:%ju be64:%ju\n", i, u, static_cast<unsigned>(htole16(u)), static_cast<unsigned>(htobe16(u)), static_cast<unsigned>(htole32(u)), static_cast<unsigned>(htobe32(u)), htole64(u), htobe64(u));
	    printf("hex = 0x%jX\n", u);
	    printf("oct = 0%jo\n", u);

	    // print binary
	    // sizeof returns the number of bytes, multiply by 8+2 for 8 bits per byte plus 2 single quote separators.
	    char buf[sizeof(intmax_t)*(8+2)+1];
	    int k = sizeof(buf)-1;
	    buf[k]=0;
	    if(u == 0) {
		buf[--k]='0';
	    } else {
		int cnt = -1; // cnt is used to insert a single quote every 4 digits
		while(u)
		{
		    if (++cnt == 4) {
			buf[--k] = '\'';
			cnt = 0;
		    }
		    buf[--k] = (u&1)?'1':'0';
		    u /= 2;
		}
	    }
	    printf("bin = 0b%s\n", buf+k);
	}

nextarg: ;
    }
    return 0;
}
